use std::collections::HashMap;
use std::path::{Path, PathBuf};

use anyhow::anyhow;
use deltalake::action::Protocol;
use deltalake::schema::SchemaTypeStruct;
use deltalake::{DeltaTableBuilder, DeltaTableMetaData};
use log::*;
use url::Url;

/*
 * Create a Delta Table with a schema
 */
pub async fn create_table_with_schema(
    output_url: Url,
    table_schema: SchemaTypeStruct,
) -> Result<(), anyhow::Error> {
    debug!("Attempting to create table in: {}", output_url.as_str());
    let mut table = DeltaTableBuilder::from_uri(output_url.as_str()).build()?;

    let protocol = Protocol {
        min_reader_version: 1,
        min_writer_version: 1,
    };

    let metadata = DeltaTableMetaData::new(None, None, None, table_schema, vec![], HashMap::new());
    table.create(metadata, protocol, None, None).await.unwrap();
    Ok(())
}

/*
 * Let's createa a Delta Table
 */
pub async fn create_table(output_url: Url, schema_path: PathBuf) -> Result<(), anyhow::Error> {
    debug!(
        "Creating table at {} with the schema from: {:?}",
        output_url, schema_path
    );

    if output_url.scheme() == "file" {
        let url = output_url.as_str().replace("file://", "");
        let path = Path::new(&url);
        debug!("Attempting to create directory path: {:?}", path);
        std::fs::create_dir_all(path)?;
    }

    let reader = std::fs::File::open(schema_path.as_path())?;
    let table_schema: SchemaTypeStruct = serde_yaml::from_reader(reader)?;
    debug!("Read YAML in: {:?}", table_schema);

    create_table_with_schema(output_url, table_schema).await
}

/*
 * Take an existing table and update
 */
pub async fn update_table_with_schema(
    output_url: Url,
    table_schema: SchemaTypeStruct,
) -> Result<(), anyhow::Error> {
    Ok(())
}

/*
 * Converts a relative table path into the full table URI based on the provided path in the options
 */
pub fn table_path_to_uri(base_uri: &str, table_path: &PathBuf) -> Result<Url, anyhow::Error> {
    debug!(
        "Converting table to URI: base: {} path: {:?}",
        base_uri, table_path
    );

    /*
     * The Url.join method treats the trailing slash as something important. In the case where it
     * doesn't exist, this should be added to ensure the join() works
     */
    let base_uri = match base_uri.chars().last() {
        Some(last) => {
            if last != '/' {
                format!("{}/", base_uri)
            } else {
                base_uri.to_string()
            }
        }
        None => base_uri.to_string(),
    };

    match table_path.as_path().to_str() {
        Some(table_path_str) => match Url::parse(&base_uri) {
            Ok(url) => Ok(url.join(table_path_str)?),
            Err(url::ParseError::RelativeUrlWithoutBase) => {
                // If we don't have a base we need to convert this to a base and try again
                let base_uri = filepath_to_url(&PathBuf::from(base_uri))?;
                println!("Foo: {} {:?}", base_uri, table_path);
                table_path_to_uri(&base_uri.into_string(), &table_path)
            }
            Err(other) => Err(other.into()),
        },
        None => Err(anyhow!("Failed to convert to a PathBuf")),
    }
}

/*
 * Convert the path of a yaml file to the path of the table in S3
 */
pub fn yamlpath_to_table(root: &str, local_path: &Path) -> Result<PathBuf, anyhow::Error> {
    let path = local_path.strip_prefix(root)?;

    if let Some(table_prefix) = path.parent() {
        match path.file_stem() {
            None => {}
            Some(table_name) => {
                return Ok(table_prefix.join(table_name));
            }
        }
    }
    Err(anyhow!(
        "Failed to convert path to table: {}",
        path.display()
    ))
}

/*
 * This function takes a file path and returns a full file:// URL
 */
pub fn filepath_to_url(path: &PathBuf) -> Result<Url, anyhow::Error> {
    match Url::from_file_path(&path) {
        Ok(url) => Ok(url),
        Err(_) => {
            // It looks like we have a relative Url
            // we'll have to build it manually!
            let pwd = std::env::var("PWD")?;
            let newp = Path::new(&pwd);
            let newpath = newp.join(path.as_path());
            filepath_to_url(&newpath)
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_path_to_table() {
        let path = Path::new("namespaces/home/pge/electric_bronze.yml");

        let table_path = yamlpath_to_table("namespaces", &path).expect("Failed to pull the path");
        assert_eq!(table_path, PathBuf::from("home/pge/electric_bronze"));
    }

    #[test]
    fn test_table_path_to_uri() {
        let path = PathBuf::from("home/pge/electric_bronze");
        let uri = table_path_to_uri("s3://delta", &path).expect("Failed to pull the URI");
        assert_eq!(
            uri,
            Url::parse("s3://delta/home/pge/electric_bronze").unwrap()
        );
    }

    #[test]
    fn test_table_path_to_uri_normalize_slashses() {
        let path = PathBuf::from("home/pge/electric_bronze");
        let uri = table_path_to_uri("s3://delta/", &path).expect("Failed to pull the URI");
        assert_eq!(
            uri,
            Url::parse("s3://delta/home/pge/electric_bronze").unwrap()
        );
    }

    #[test]
    fn test_table_path_to_uri_with_local_base() {
        let _ = pretty_env_logger::try_init();
        let result = table_path_to_uri("output", &PathBuf::from("relative"));
        assert!(result.is_ok());
        let result = result.unwrap();

        let pattern = "output/relative";
        // Make sure that the output matches the pattern we expect
        // Doing this with a matches() since the root the result will be dependent on the test
        // runner's location
        assert_eq!(
            1,
            result.path().matches(pattern).collect::<Vec<&str>>().len(),
            "result was: {}",
            result
        );
    }

    #[test]
    fn test_path_to_uri_relative() {
        let simple = filepath_to_url(&PathBuf::from("relative"));
        assert!(simple.is_ok());
        let simple = simple.unwrap();
        assert_eq!(simple.scheme(), "file");
    }

    #[test]
    fn test_path_to_uri_relative_slash() {
        let simple = filepath_to_url(&PathBuf::from("./relative"));
        assert!(simple.is_ok());
        let simple = simple.unwrap();
        assert_eq!(simple.scheme(), "file");
    }
}
