/*
 * This file contains the schema migration/creation tool
 */

use dotenv::dotenv;
use gumdrop::Options;
use log::*;

#[derive(Debug, Options)]
struct CLIOptions {
    #[options(help = "Print this very helpful message")]
    help: bool,
    #[options(help = "Path which contains yaml table definitions", required)]
    from: String,
    #[options(help = "Destination path to create/migrate the tables", required)]
    to: String,
}

#[async_std::main]
async fn main() -> Result<(), anyhow::Error> {
    dotenv().ok();
    match std::env::var("RUST_LOG") {
        Ok(_) => {}
        Err(_) => {
            // Default to at least warn level logging
            std::env::set_var("RUST_LOG", "warn");
        }
    }
    pretty_env_logger::init();

    let opts = CLIOptions::parse_args_default_or_exit();
    debug!("Starting with options: {:?}", opts);

    use glob::glob;
    let pattern = format!("{}/**/*.yml", opts.from);
    debug!("Using glob pattern: {}", pattern);

    for entry in glob(&pattern).expect("Failed to read glob pattern") {
        match entry {
            Ok(path) => {
                debug!("Considering: {:?}", path);
                match carto::yamlpath_to_table(&opts.from, &path) {
                    Ok(table_path) => {
                        let url = carto::table_path_to_uri(&opts.to, &table_path)
                            .expect("Failed to create the table URI");
                        debug!("Attempting to open table at: {}", url);

                        match deltalake::open_table(url.clone()).await {
                            Ok(table) => {
                                debug!("table: {:?}", table);
                            }
                            Err(deltalake::DeltaTableError::NotATable(st)) => {
                                warn!("Table does not exist {}", st);
                                carto::create_table(url, path).await?;
                            }
                            Err(e) => {
                                error!("Failed to open table: {:?}", e);
                            }
                        }
                    }
                    Err(e) => {
                        error!("Failed to use local path to compute table path: {:?}", e);
                    }
                }
            }
            Err(e) => error!("{:?}", e),
        }
    }
    Ok(())
}
